//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%
#include <CCfits/CCfits>
#include <Model/Component/Component.h>
#include <XSModel/Parameter/TableModParam.h>

// UniqueEnergy
#include <XSModel/Model/UniqueEnergy.h>
// ModParam
#include <XSModel/Parameter/ModParam.h>
// OGIPTable
#include <XSModel/Model/Component/OGIPTable/OGIPTable.h>

#include <XSFunctions/Utilities/FunctionUtility.h>
#include <Model/Component/AddTableComponent.h>
#include <Model/Component/MulTableComponent.h>
#include <XSUtil/Numerics/LinearInterp.h>
#include <XSUtil/Utils/XSstream.h>

#include "SPutils.h"

#include <XSstreams.h>
#include <iostream>
#include <memory>
#include <sstream>
#include <functional>


// Class OGIPTable 
const Real OGIPTable::FUZZY = 1.E-06;
const long OGIPTable::s_MEM_THRESHOLD = 10000000;

OGIPTable::OGIPTable(const OGIPTable &right)
      : TableComponent(right),
	m_table(right.m_table),
        m_interParam(),
        m_startWeight(right.m_startWeight),
        m_endWeight(right.m_endWeight),
        m_startWeightBin(right.m_startWeightBin),
        m_endWeightBin(right.m_endWeightBin),
        m_addParam()
{
   // m_interParam and m_addParam will be set at a 
   // later point either in the setParamPointersFromCopy or read function.
}

OGIPTable::OGIPTable (const string& nameString, Component* p)
      : TableComponent(nameString,p),
	m_table(),
        m_interParam(),
        m_startWeight(),
        m_endWeight(),
        m_startWeightBin(),
        m_endWeightBin(),
        m_addParam()
{
}


OGIPTable::~OGIPTable()
{
}


void OGIPTable::read (bool readSpectralData, Component* p, bool isAddRequest)
{

  // get the file size to see whether we want to read in all the data from
  // the SPECTRA extension at the start or just read it when required.

  bool readAll(false);
  struct stat fileStat;
  if ( stat(filename().c_str(), &fileStat) != 0 ) {
    string diag = "Cannot find table model file " + filename();
    throw YellowAlert(diag);
  } else {
    if ( fileStat.st_size < s_MEM_THRESHOLD ) readAll = true;
  }

  // read in the table, reading the data from the SPECTRA extension if the
  // file is small enough
  
  int status = m_table.read(filename(), readAll);
  if ( status != heasp::OK ) {
    string diag(filename());
    diag += " : " + heasp::SPgetErrorStack() + "\n";
    throw YellowAlert(diag);       
  }

  bool isAdd = m_table.getisAdditive();

  // Owning component pointer p will be NULL if this is being
  // called from outside normal xspec channels - ie. if using
  // TableModel wrapper functions.  In that case check the
  // isAddRequest flag instead.
  if (p) {
    if ( isAdd && !dynamic_cast<AddTableComponent*>(p) ) {
      string diag = " multiplicative model requested \nbut file ";
      diag += filename();
      diag += " contains additive model";    
      throw WrongTableType(diag);   
    } else if (!isAdd && dynamic_cast<AddTableComponent*>(p) ) {
      string diag = " additive model requested \nbut file ";
      diag += filename();
      diag += " contains multiplicative model";    
      throw WrongTableType(diag);   
    }
    // check whether the table uses xflt and if so set the spectrumDependency
    // in the parent Component object
    if ( m_table.getFiltExps().size() > 0 ) p->setSpectrumDependency(true);
  } else {
    if (isAdd && !isAddRequest) {
      string diag = " multiplicative model requested \nbut file ";
      diag += filename();
      diag += " contains additive model";    
      throw WrongTableType(diag);   
    } else if (!isAdd && isAddRequest) {
      string diag = " additive model requested \nbut file ";
      diag += filename();
      diag += " contains multiplicative model";    
      throw WrongTableType(diag);   
    }          
  }

  parent(p);

  bool redshift = m_table.getisRedshift();
  bool escale = m_table.getisEscale();
  string modName = m_table.getModelName(); 
  if (modName.length() < 2) {
    string diag("Table Model MODLNAME keyword must contain a name with at least 2 characters.\n");
    throw YellowAlert(diag);
  }

  name(modName);

  // set parameter info

  int nInterpParams(m_table.getNumIntParams());
  int nAdditionalParams(m_table.getNumAddParams());
  int Nparams = nInterpParams + nAdditionalParams;

  numIntPar(nInterpParams);
  numAddPar(nAdditionalParams);

  if (!p) {
    // If there's no owning component -- meaning this is coming from a
    // TableModel wrapper function, then we have 2 extra things to
    // be aware of:
    //   1.  Can't assume m_interParam, m_addParam, and TableComponent's
    //       m_params are all empty, since this object is persistent
    //       and may go through many read calls.  Therefore we must clear
    //       those arrays here.
    //   2.  TableComponent's m_params is the OWNER of these pointers,
    //       since there is no parent Component and these are not registered
    //       in XSPEC's global containers.
    m_interParam.clear();
    m_addParam.clear();
    std::vector<Parameter*>& owningPointers = params();
    for (size_t i=0; i<owningPointers.size(); ++i)
      delete owningPointers[i];
    owningPointers.clear();
  }

  // the following loop works both for the cases of all the interpolation
  // parameters followed by all the additive parameters or them mixed 
  // amongst each other with additive parameters indicated by method=-1 or
  // no tabulated parameter values.

  for (int i = 0; i < Nparams; ++i) {

    heasp::tableParameter thisParam(m_table.getParametersElement(i));

    int numbVals = thisParam.getNumberTabulatedValues();

    // If the table parameter has no unit string, heasp may return
    //  a blank but non-empty string.  Remove leading/trailing blanks here.
    string unitStr;
    string::size_type startPos = thisParam.getUnits().find_first_not_of(' ');
    string::size_type endPos = thisParam.getUnits().find_last_not_of(' ');
    if (startPos != string::npos && endPos != string::npos)
       unitStr = thisParam.getUnits().substr(startPos,endPos-startPos+1);
           
    if (thisParam.getInterpolationMethod() >= 0 && 
	m_interParam.size() < (size_t)nInterpParams && numbVals > 0) {
      // case of an interpolated parameter
      std::unique_ptr<TableModParam>  modParam 
	(new TableModParam(thisParam.getName(), 0, 
			   thisParam.getInitialValue(), thisParam.getDelta(), 
			   thisParam.getMaximum(), thisParam.getMinimum(), 
			   thisParam.getTop(), thisParam.getBottom(), 
			   unitStr));

      modParam->numVals(numbVals);
      modParam->setTabValue(thisParam.getTabulatedValues());
      modParam->logInterp(thisParam.getInterpolationMethod());
      params().push_back(modParam.release());
      m_interParam.push_back(static_cast<TableModParam*>(params().back()));
    } else {
      // case of an additive parameter
      std::unique_ptr<ModParam> modPar
	(new ModParam(thisParam.getName(), 0, 
		      thisParam.getInitialValue(), thisParam.getDelta(), 
		      thisParam.getMaximum(), thisParam.getMinimum(), 
		      thisParam.getTop(), thisParam.getBottom(), 
		      unitStr));
      params().push_back(modPar.release());
      m_addParam.push_back(static_cast<ModParam*>(params().back()));
    }
  }

  if ( escale ) {
    // construct a new modparam 
    params().push_back(new ModParam(string("Escale"),0,1.,-0.001,100.,0.00001,
				    100.,0.00001));
  }

  if ( redshift ) {
    // construct a new modparam 
    params().push_back(new ModParam(string("z"),0,0.,-0.001,5.,0.,5.,0.));
  }

}

OGIPTable* OGIPTable::clone (Component* parent) const
{
   OGIPTable* newTable = new OGIPTable(*this);
   newTable->parent(parent);
   newTable->setParamPointersFromCopy();
   return newTable;
}

bool OGIPTable::formatCheck (const string& fileName)
{
  // check that this is a valid format file by trying to read it
  heasp::table testTable;
  int status = testTable.read(fileName, false);
  if ( status == heasp::OK ) {
    return true;
  } else {
    return false;
  }
}

void OGIPTable::getInterpolant (const Real minEnergy, const Real maxEnergy,
				const std::map<string,Real>& xfltInfo,
				RealArray& spectrum, RealArray& variance)
{

  // construct array of current parameter values

  std::vector<Parameter*>& parameters(params());
  RealArray parameterValues(parameters.size());
  for (size_t i=0; i<parameterValues.size(); i++) {
    parameterValues[i] = parameters[i]->value();
  }

  // use table class method to get interpolated values

  RealArray tableEnergyBins;
  int status = m_table.getValues(parameterValues, xfltInfo, minEnergy, 
				 maxEnergy, tableEnergyBins, spectrum, 
				 variance);
  if ( status == heasp::InconsistentNumTableParams ) {
    string err("Inconsistent number of table parameters detected. This should not happen.");
    throw YellowAlert(err);
  } else if ( status == heasp::InconsistentTableFilter ) {
    string err("Inconsistent XFLT keywords between datasets and table model file.");
    throw YellowAlert(err);
  } else if ( status == heasp::TableParamValueOutsideRange ) {
    string err("Attempt to access parameter value outside tabulated range. Inconsistency in table model file.");
    throw YellowAlert(err);
  }

}

void OGIPTable::energyWeights (const UniqueEnergy* uniqueEng)
{
  // produce weights for elements of the energy array. 
  // these are used to rebin the componentSpectrum - which is on the
  // energy grid of the tabulated data - into the energy array of the model
  // component.   

  // save the last values of the target energies, redshift, and Escale
  // (the latter two if necessary) to determine whether the weights need
  // to be recalculated

/*  static std::map<const UniqueEnergy*,RealArray> savedEnergyArray;
  static std::map<const UniqueEnergy*,Real> savedRedshift;
  static std::map<const UniqueEnergy*,Real> savedEscale;
*/
  using namespace std::placeholders;
  // the target energy array
  const RealArray& energyArray = uniqueEng->energy();
  size_t Nout ( energyArray.size() - 1);

  // references to the weights

  IntegerVector& startBin = m_startWeightBin[uniqueEng];
  IntegerVector& endBin   = m_endWeightBin[uniqueEng];        
  RealArray& startWeight = m_startWeight[uniqueEng];
  RealArray& endWeight   = m_endWeight[uniqueEng];                

  // check for a z parameter

  XSutility::MatchPtrName<Parameter> matchName;
  std::vector<Parameter*>::iterator r =
                   std::find_if(params().begin(),params().end(),std::bind(matchName,_1,"z"));
  Real z (0);
  if ( r != params().end() ) z = (*r)->value();
  if (z <= -1.0) {
     string err("Cannot calculate table model with redshift z <= -1.0\n");
     throw YellowAlert(err);
  }
  Real zf = 1.0/(1.0+z);

  // check for an Escale parameter

  r = std::find_if(params().begin(),params().end(),std::bind(matchName,_1,"Escale"));
  Real energyScaling (1.0);
  if ( r != params().end() ) energyScaling = (*r)->value();
  if (energyScaling <= 0.0) {
     string err("Cannot calculate table model with energy scaling Escale <= 0.0\n");
     throw YellowAlert(err);
  }

  // check whether the weights need to be recalculated and if not return

/*  bool recalc = false;
  if ( (z != savedRedshift[uniqueEng]) || 
       (energyScaling != savedEscale[uniqueEng]) ||
       energyArray.size() != savedEnergyArray[uniqueEng].size()) recalc = true;
  if ( !recalc ) {
    RealArray& savedEnergies = savedEnergyArray[uniqueEng];
    for (size_t i=0; i<savedEnergies.size(); i++) {
      if ( energyArray[i] != savedEnergies[i] ) {
	recalc = true;
	break;
      }
    }
  }
  if ( recalc ) {
    savedRedshift[uniqueEng] = z;
    savedEscale[uniqueEng] = energyScaling;
    savedEnergyArray[uniqueEng] = energyArray;
  } else {
    return;
  }
*/
  // modify the table energies by redshift and energy scaling 

  const std::vector<Real>& tabEnergies = m_table.getEnergies();
  RealArray observedEnergy(tabEnergies.size());
  for (size_t k = 0; k < observedEnergy.size(); ++k) observedEnergy[k] = tabEnergies[k]*zf*energyScaling;

  size_t inputBin = 0;
  size_t outputBin = 0;

  // DO NOT stream output to tcout.  This code may be accessed through xsmtbl and xsatbl 
  // by users linking the models library to their own programs. 


  if (!Numerics::Rebin::findFirstBins(observedEnergy, energyArray, FUZZY, 
                inputBin, outputBin))
  {
     // Deliberately not throwing here for (at least) 2 reasons:
     // Data command doesn't expect its Notify call to throw, and therefore
     // isn't catching anything at that point.  Also it's still possible
     // a calculation will be performed on this model, in which case it
     // expects to have valid bin and weight arrays (doing "plot model"
     // after throwing from here caused a seg-fault).
     std::ostringstream oss;  
     oss << "\n There is no model information available in the requested observed frame "
	  << " energy range ( " << energyArray[0] << "," << energyArray[Nout] << ")\n"
	 << " the model rest frame energy range is " << observedEnergy[0] << " to " << observedEnergy[observedEnergy.size()-1] << '\n';
     if ( z != 0.0 ) oss << " redshift requested is " << z << '\n';
     if ( energyScaling != 1.0 ) oss << " energy scaling requested is " << energyScaling << '\n';
     FunctionUtility::xsWrite(oss.str(),10);
     size_t Nout (energyArray.size() - 1);
     startWeight.resize(Nout,0);
     endWeight.resize(Nout,0);
     startBin.clear();
     endBin.clear();
     startBin.resize(Nout,0);
     endBin.resize(Nout,0);
  }
  else
  {
     if (outputBin > 0) 
     {
       std::ostringstream oss;
       oss << "\n No model information is available for observed frame energies below "
	     << std::setw(14) << observedEnergy[0] 
	     << " so model is set to zero at these energies\n";
       FunctionUtility::xsWrite(oss.str(),15);
     }
     
     Numerics::Rebin::initializeBins(observedEnergy, energyArray, FUZZY, inputBin, outputBin,
                            startBin, endBin, startWeight, endWeight);
     if ( inputBin >= observedEnergy.size()-1 ) {
       std::ostringstream oss;
       oss << "\nNo model information is available for observed frame energies \n"
	   << "above " << observedEnergy[observedEnergy.size()-1] 
	   << " so model is taken as zero in that range\n";
       FunctionUtility::xsWrite(oss.str(),15);
     }
  }

}

void OGIPTable::clearArrays (const std::set<UniqueEnergy*>& currentEngs)
{

  std::map<const UniqueEnergy*,IntegerVector>::iterator m (m_startWeightBin.begin());
  std::map<const UniqueEnergy*,IntegerVector>::iterator mEnd (m_startWeightBin.end());

  while ( m != mEnd )
  {
     // Casting away a const is generally not a good thing, but
     // it's only to allow m->first to be used as a key.
     // It doesn't actually modify what nm points to.
       UniqueEnergy* nm = const_cast<UniqueEnergy*>(m->first);
       ++m;
       if (currentEngs.find(nm) == currentEngs.end())
       {
           m_startWeightBin.erase(nm);           
           m_endWeightBin.erase(nm);           
           m_startWeight.erase(nm);           
           m_endWeight.erase(nm);           
       }         
  }
}

void OGIPTable::rebinComponent (const UniqueEnergy* uniqueEng, const RealArray& inputArray, RealArray& outputArray, bool variance)
{

  IntegerVector& startBin   = m_startWeightBin[uniqueEng];     
  IntegerVector& endBin     = m_endWeightBin[uniqueEng];     
  RealArray endWeight (m_endWeight[uniqueEng]);     
  RealArray startWeight (m_startWeight[uniqueEng]);     

  if ( variance )
  {
        startWeight *= m_startWeight[uniqueEng];
        endWeight   *= m_endWeight[uniqueEng];       
  }

  Real lowValue = m_table.getLowEnergyLimit();
  Real highValue = m_table.getHighEnergyLimit();
  if ( lowValue < 0.0 ) lowValue = 0.0;
  if ( highValue < 0.0 ) highValue = 0.0;

  Numerics::Rebin::rebin(inputArray, startBin, endBin, startWeight, endWeight, outputArray, lowValue, highValue);

}

void OGIPTable::interpolateComponent (const UniqueEnergy* uniqueEng, const RealArray& inputArray, RealArray& outputArray, bool exponential)
{
  IntegerVector& startBin   = m_startWeightBin[uniqueEng];     
  IntegerVector& endBin     = m_endWeightBin[uniqueEng];     
  RealArray& endWeight     = m_endWeight[uniqueEng];     
  RealArray& startWeight   = m_startWeight[uniqueEng];     

  Real lowValue = m_table.getLowEnergyLimit();
  Real highValue = m_table.getHighEnergyLimit();
  if ( lowValue < 0.0 ) lowValue = 1.0;
  if ( highValue < 0.0 ) highValue = 1.0;

  Numerics::Rebin::interpolate(inputArray, startBin, endBin, startWeight, 
			       endWeight, outputArray, exponential, lowValue, highValue);
}

void OGIPTable::setParamPointersFromCopy ()
{
   // If TableComponent's parent hasn't been set yet (ie. if copying
   // from a prototype), nothing will be done in base class
   // setParamPointersFromCopy, and by extension nothing will be done
   // in here.  This is for copying active model objects, ie. when
   // adding data in a new data group.
   TableComponent::setParamPointersFromCopy();
   std::vector<Parameter*>::const_iterator itPar = params().begin();
   std::vector<Parameter*>::const_iterator itEnd = params().end();
   while (itPar != itEnd)
   {
      // I wish there was a better way to do this.
      TableModParam* tablePar = dynamic_cast<TableModParam*>(*itPar);
      if (tablePar)
      {
         m_interParam.push_back(tablePar);
      }
      else
      {
         m_addParam.push_back(static_cast<ModParam*>(*itPar));
      }
      ++itPar;
   }
}

// Additional Declarations
